/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Christophe Ney
 * Contributor(s): France Telecom R&D
 *
 * --------------------------------------------------------------------------
 * $Id: EjbqlQueryTreeHolder.java,v 1.14 2003/08/28 12:56:42 joaninh Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas_ejb.lib;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;

import org.objectweb.jonas_ejb.deployment.api.DeploymentCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.EntityCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.MethodCmp2Desc;
import org.objectweb.jonas_ejb.deployment.ejbql.ASTEJBQL;
import org.objectweb.jonas_ejb.container.TraceEjb;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.eval.api.QueryEvaluator;
import org.objectweb.medor.eval.lib.MedorEvaluator;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.optim.api.QueryTransformer;
import org.objectweb.medor.query.api.OrderField;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.jorm.api.JormExtent;
import org.objectweb.medor.query.jorm.lib.JormQueryTreeHelper;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.lib.SelectProject;
import org.objectweb.util.monolog.api.BasicLevel;


/**
 * Class to hold the query tree of a given EJBQL request. This allows walking the
 * JormExtents of the tree to set the mappers.
 * @author  Christophe Ney - cney@batisseurs.com
 * contributor(s):
 * - Helene Joanin: Take into account the ORDER BY clause.
 */
public class EjbqlQueryTreeHolder {

    QueryTree queryTree = null;
    QueryTransformer queryOptimizer = null;
    QueryEvaluator optimizedQuery = null;
    Field resField = null;

    private String beanName;
    private String methodName;


    /**
     * constructor of a holder for a specific query.
     * @param mDesc JOnAS meta-information for the corresponding finder/select method
     * @param ejbql root of the lexical tree of the query
     * @param mapper mapper to associate at each leaves of the QueryTree.
     *               The mapper may be null in case of the QueryTree is build in the GenIC phase.
     */
    public EjbqlQueryTreeHolder(MethodCmp2Desc mDesc, ASTEJBQL ejbql, PMapper mapper)
        throws Exception {

        EntityCmp2Desc eDesc = (EntityCmp2Desc) mDesc.getBeanDesc();
        DeploymentCmp2Desc dd = (DeploymentCmp2Desc) eDesc.getDeploymentDesc();
        Manager jormManager = dd.getJormManager();

        // create identifiers Map (visit ALL the query)
        Map fields = (new EjbqlVariableVisitor(ejbql, dd)).getFields();

        // create SelectProject (visit the SELECT clause)
        SelectProject sp = new EjbqlSelectVisitor(ejbql, fields).getSelectProject();

        // create the Filter's expression (visit the WHERE clause)
        Expression qf = new EjbqlQueryFilterVisitor(mapper, fields, mDesc.getMethod().getParameterTypes(), ejbql).getQueryFilter();
        sp.setQueryFilter(qf);

        // create the OrderField list (visit the ORDER BY clause)
        OrderField[] ofs = new EjbqlOrderByVisitor(ejbql, fields).getOrderFields();
        if (ofs.length > 0) {
            sp.setOrderBy(ofs);
        }

        queryTree = sp;

        // set the mapper of each leaves of the QueryTree
        if (mapper != null) {
            Collection extents = JormQueryTreeHelper.getJormExtents(queryTree);
            for (Iterator it = extents.iterator(); it.hasNext();) {
                JormExtent je = (JormExtent) it.next();
                je.setPMapper(mapper, RdbMappingBuilder.getProjectName());
            }
        }

        // Get the field description of the result of the query
        // Nota Bene: TupleStructures of queries non optimized and optimized are differents.
        resField = queryTree.getTupleStructure().getField(1);

        // Just needed later for debug trace
        beanName = eDesc.getJormClassName();
        methodName = mDesc.getMethod().getName();
    }

    /**
     * get the query tree that was built from visiting the lexical tree
     */
    public QueryTree getQueryTree() {
        return queryTree;
    }

    /**
     * Set the optimizer to be used when optimizing the query tree
     */
    public void setQueryOptimizer(QueryTransformer qtf) {
        optimizedQuery = null;
        queryOptimizer = qtf;
    }

    /**
     * get the current query tree optimizer
     */
    public QueryTransformer getQueryOptimizer() {
        return queryOptimizer;
    }

    /**
     * get the Medor result Field of the query (useful for the ejbSelect method)
     */
    public Field getResField() {
        return resField;
    }

    /**
     * Get the query evaluator of the optimized query. The query is optimized on first call.
     * All mappers of the query should have been already set with the help of the iterator
     * returned by the <code>getJormExtentIterator</code> method.
     */
    public QueryEvaluator getOptimizedQueryTree() throws Exception {
        if (optimizedQuery == null) {
            QueryTreePrinter.printQueryTree(queryTree, TraceEjb.query, BasicLevel.DEBUG);
            QueryTree qt = queryOptimizer.transform(queryTree);
            TraceEjb.query.log(BasicLevel.DEBUG,
                               "Optimize the QueryTree of the method " + methodName
                               + " of the bean " + beanName);
            QueryTreePrinter.printQueryTree(qt, TraceEjb.query, BasicLevel.DEBUG);
            optimizedQuery = new MedorEvaluator(qt, 0);
        }
        return optimizedQuery;
    }

}
