/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Xavier Delplanque & Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: AbsConfigGenerator.java,v 1.7 2003/09/19 15:07:47 delplanx Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas_ws.wsgen;

import org.objectweb.jonas_ejb.deployment.api.BeanDesc;

import org.objectweb.jonas_ws.deployment.api.WSDeploymentDesc;
import org.objectweb.jonas_ws.deployment.api.ServiceRef;
import org.objectweb.jonas_ws.deployment.api.ServiceDesc;
import org.objectweb.jonas_ws.deployment.api.ServiceRef;

import org.objectweb.jonas.common.Log;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.List;
import java.util.Map;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Iterator;
import java.util.Enumeration;

import java.net.URL;
import java.net.URLClassLoader;

import java.io.File;
import java.io.FileFilter;

import org.objectweb.common.Cmd;

/**
 * This class realizes configuration files generation.
 * This class implements genEndpointConfig & genClientConfig methods;
 * configuration methods stay abstract (WS implementation specific).
 */
public abstract class AbsConfigGenerator implements ConfigGenerator {

    /** WsGen logger */
    protected static Logger logger = Log.getLogger(Log.JONAS_WSGEN_PREFIX);

    /** The output directory where files are generated */
    protected File genDir = null;

    /** The $JAVA_HOME/bin directory */
    protected String javaHomeBin = null;

    /** The classpath to use with command lines */
    protected String cp = "";

    /** The complete name of the file processed*/
    protected String filename = "";
    
    private Map options = new Hashtable(); 
    /**
     * Constructor : creates a AbsConfigGenerator object.
     */
    public AbsConfigGenerator() {
	
	// get JAVA_HOME environment property.
	// used for compilation and file generation.
        if (javaHomeBin == null) {
            javaHomeBin = System.getProperty("java.home", "");
            if (!("".equals(javaHomeBin))) {
                javaHomeBin = javaHomeBin + File.separator + ".."
                    + File.separator + "bin" + File.separator;
            }
        }
	
	// transform current class loader in classpath.
	cp = getContextClassPath();

    }

    public String getDefaultMapping() {
	return "/services/*";
    }

    public void genEndpointConfig(WSDeploymentDesc dd,
				  String filename)
	throws WsGenException  {

        if (logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "Endpoints generation in " + filename);

	// create a temp directory for file generation
	try {
	    genDir = WsGen.createTempDir();
	}
	catch (java.io.IOException ie) {
	    throw new WsGenException("Cannot create the temporary directory for endpoint config file generation",ie);
	}

	// generate files in the temp directory
	configureEndpoint(dd);
	
	// create a manager for adding files in the given jar
	JarManager jm = new JarManager(filename);
	jm.addDir(genDir);
	jm.close(); 

	// remove the temp directory
	destroyTempDir();
    }
    
    public class JavaFileFilter implements FileFilter {
	public JavaFileFilter() {
	}

	public boolean accept(File file) {
	    return file.getName().endsWith(".java");
	}
    }


    public class DirFileFilter implements FileFilter {
	public DirFileFilter() {
	}

	public boolean accept(File file) {
	    return file.isDirectory();
	}
    }

    protected void addSrcs (File dir, Vector srcs)
	throws WsGenException {
	
	try {
	    
	    // add each java sources contained in the directory
	    File[] files = dir.listFiles(new JavaFileFilter());	 
	    for (int i = 0; i < files.length; i++) {
		srcs.add(files[i]);    
	    }
	    
	    // reapply on subdirectories
	    files = dir.listFiles(new DirFileFilter());	 
	    for (int i = 0; i < files.length; i++) {
		addSrcs(files[i],srcs);
	    }
	    
	} catch (Exception e) {
	    throw new WsGenException("Error when adding files for directory '"+dir.getName()+"'",e);
	}

    }
    
    protected void compileGenDir()
	throws WsGenException  {

	try {
	    // get all java files contained in the gen dor
	    Vector srcs = new Vector();
	    addSrcs(genDir,srcs);
	    // todo : define if we keep generated sources	    
	    // todo : get the javac command & others options
	    Cmd cmd = new Cmd(javaHomeBin+"javac");
	    cmd.addArgument("-classpath");
	    cmd.addArgument(cp);
	    cmd.addArgument("-d");
	    cmd.addArgument(genDir.getCanonicalPath());
	    // cmd.addArguments(optionsJavaC);
	    if (srcs.size() != 0) {
		for (Enumeration e = srcs.elements(); e.hasMoreElements();) {
		    String srcName = ((File)e.nextElement()).getCanonicalPath();
		    cmd.addArgument(srcName);
		}
		trace("Running '" + cmd.toString() + "'");	    
		if (cmd.run()) {
		    WsGen.info("Web Services Classes successfully compiled.");
		}
		else {
		    String err = "Error when running javac tool.";
		    throw new WsGenException(err);
		}
		
		// destroy java files if keep generated option is not set
		boolean keepGen = ((Boolean)getOption(KEEP_GEN)).booleanValue();
		if (!keepGen) {
		    for (Enumeration e = srcs.elements(); e.hasMoreElements();) {
			File srcFile = (File)e.nextElement();
			srcFile.delete();
		    }		
		}
	    }
	}
	catch (Exception e) {
	    throw new WsGenException("Error when compiling sources",e);	    
	}


    }

    protected void genClientFiles(ServiceRef[] servRefs)
	throws WsGenException  {
        for (int i = 0 ; i < servRefs.length ; i++ ) {
	    ServiceRef sr = servRefs[i];
	    configureClient(sr);
	}
    }
    /**
     * Client config generation for ajbjar file
     */    
    public void genClientConfig(Iterator bdIt,
				String filename)
	throws WsGenException  {

        if (logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "Clients generation in " + filename);

	// create a temp directory for file generation
	try {
	    genDir = WsGen.createTempDir();
	}
	catch (java.io.IOException ie) {
	    throw new WsGenException("Cannot create the temporary directory for client config file generation",ie);	}
	this.filename = filename;

	while(bdIt.hasNext()) {
	    BeanDesc bd = (BeanDesc)bdIt.next();
	    if (bd.getServiceRefDesc().length != 0) {
		// the ejb may have clients
		// add clients files in temporary file
		genClientFiles(bd.getServiceRefDesc());
	    }		
	}

	compileGenDir();

	// create a manager for adding files in the given jar
	JarManager jm = new JarManager(filename);
	jm.addDir(genDir);
	System.out.println ("add files in the jar ... ");
	jm.close();
	System.out.println ("... files added");

	// remove the temp directory
	destroyTempDir();	
    }
    
    /**
     * Client config generation for war file
     */
    public void genClientConfig(ServiceRef[] servRefs,
				String filename)
	throws WsGenException  {

        if (logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "Clients generation in " + filename);

	// create a temp directory for file generation
	try {
	    genDir = WsGen.createTempDir();
	}
	catch (java.io.IOException ie) {
	    throw new WsGenException("Cannot create the temporary directory for client config file generation",ie);	
	}

	this.filename = filename;
	
	genClientFiles(servRefs);
       	compileGenDir();

	// create a manager for adding files in the given jar
	WarManager wm = new WarManager(filename);
	
	// update service interfaces
	for (int i = 0; i < servRefs.length; i++) {
	    ServiceRef sr = servRefs[i];
	    if (!sr.getServiceInterface().getName().equals("javax.xml.rpc.Service")) {
		// replace this interface by the generated one
		// todo : 
		//wm.setSEI(sr.getServiceRefName(), getSEIName(sr));
	    }
	}

	wm.addDir(genDir);
	System.out.println ("add files in the war ... ");
	wm.close();
	System.out.println ("... files added");

	// remove the temp directory
	destroyTempDir();
    }
    
    /**
     * Recursively destroy the given file.
     * @param file the file to destroy.
     */
    protected boolean destroyFile (File file) {
	boolean isDestroy = true;
	if (file.isDirectory()) {
	    // destroy all childs	    
	    File[] childs = file.listFiles();
	    for (int i = 0; i < childs.length; i++) {
		isDestroy &= destroyFile(childs[i]);
	    }
	}
	return isDestroy && file.delete();
    }
    
    /**
     * Destroy the temporary directory used for generation.
     *
     */
    protected void destroyTempDir(){
	if (!destroyFile(genDir)){
	    WsGen.warning("Cannot delete temporary dir '"+genDir.getName()+"'");
	}	
    }
    
    /**
     * Configure Endpoint has to be implemented in subClass.
     * hold specific process for endpoint with specific WebService engine.
     * Generate files in the temporary directory genDir.
     * Needs to be overridden in sub classes.
     *
     * @param wsdd the webservices deployment descriptor
     *
     * @throws WsGenException if generation fails
     *                        if the method isn't implemented
     */
    protected void configureEndpoint(WSDeploymentDesc wsdd)
	throws WsGenException {
	throw new WsGenException("configureEndpoint needs to be overridden " +
				 "in subclasses of " + this.getClass().getName());
    }

    /**
     * ConfigureClient has to be implemented in subClass.
     * hold specific process for client with specific WebService engine.
     * Generate files in the temporary directory genDir.
     * Needs to be overridden in sub classes.
     *
     * @param sr The ServiceRef representing the client dependency.
     *
     * @throws WsGenException if generation fails
     *                        if the method isn't implemented
     */
    protected void configureClient(ServiceRef sr)
	throws WsGenException {
	throw new WsGenException("configureClient needs to be overridden " +
				 "in subclasses of " + this.getClass().getName());
    }
    
    /**
     * Return the webservices engine class name.
     * @return the webservices engine class name.
     *
     * @throws WsGenException if the method isn't implemented
     */
    public String getEngineWebApp()
	throws WsGenException {
	throw new WsGenException("getEngineWebApp needs to be overridden " +
				 "in subclasses of " + this.getClass().getName());    
    }


    /**
     * Return a String containing the path representation
     * of the ContextClassLoader.
     *
     * @return Return a String containing the path representation
     * of the ContextClassLoader.
     */
    private static String getContextClassPath() {
	URLClassLoader ucl = (URLClassLoader)(Thread.currentThread().getContextClassLoader());
	String ps = System.getProperty("path.separator");
	String cp = ".";
        URL[] urls = ucl.getURLs();
        for (int i = 0; i < urls.length; i++) {
            cp = cp + ps + urls[i].getFile();
        }
        return cp;
    }

    /**
     * Display the specified message only if verbose.
     * @param msg the message to display
     */
    protected void trace(String msg) {
        logger.log(isVerbose() ? BasicLevel.INFO : BasicLevel.DEBUG, msg);
    }

    /**
     * return true if verbose mode is set.
     * @return true if verbose mode is set.
     */
    protected boolean isVerbose () {
	return ((Boolean)getOption(VERBOSE, VERBOSE_DEFAULT)).booleanValue();
    }

    /**
     * return true if debug mode is set.
     * @return true if debug mode is set.
     */
    protected boolean isDebug () {
	return ((Boolean)getOption(DEBUG, DEBUG_DEFAULT)).booleanValue();
    }

    /**
     * return true if noConfig mode is set.
     * @return true if noConfig mode is set.
     */
    protected boolean isNoConfig () {
	return ((Boolean)getOption(NO_CONFIG, NO_CONFIG_DEFAULT)).booleanValue();
    }

    /**
     * Set an option for generation.
     * @param name option name
     * @param value option value
     */
    public void setOption(String name, Object value){
	options.put(name, value);
    }

    /**
     * Get an option, return the option value or default
     * if the option isn't set.
     * @param name option name
     * @param defolt default option value
     *
     * @return the option value or default
     * if the option isn't set.
     */
    public Object getOption(String name, Object defolt) {
	Object res = getOption(name);
	if (res == null){
	    res = defolt;
	}
	return res;
    }

    /**
     * Get an option, return the option value or null
     * if the option isn't set.
     * @param name option name
     *
     * @return the option value or null
     * if the option isn't set.
     */
    public Object getOption(String name) {
	return options.get(name);
    }

}
